
include "tools.pyx"

cdef extern from "Python.h":
    void Py_INCREF(object)
    void Py_DECREF(object)
    double PyFloat_AsDouble(object)
    char *PyString_AsString(object)

cdef extern from "string.h":
   char *strcpy(char *dest, char *src)
   char *strncpy(char *dest, char *src, size_t n)

cdef extern from "stdlib.h":
     void* malloc(size_t size)
     void free(void* ptr)
     void* mcpy(void* ptr1, void* ptr2, size_t size)

cdef ptr_wrapper(void *ptr):
    return PyCObject_FromVoidPtr(ptr, NULL)

cdef void *ptr_unwrapper(object obj):
    return PyCObject_AsVoidPtr(obj)


#/*****************************************
#*       ABSTRACT DATA TYPES              *
#*****************************************/
PGA_DATATYPE_BINARY=1
PGA_DATATYPE_INTEGER=2
PGA_DATATYPE_REAL=3
PGA_DATATYPE_CHARACTER=4
PGA_DATATYPE_USER=5

ctypedef unsigned long PGABinary
ctypedef signed long int PGAInteger
ctypedef double PGAReal
ctypedef signed char PGACharacter

PGA_INT=1
PGA_DOUBLE=2
PGA_CHAR=3
PGA_VOID=4


#/*****************************************
#*              BOOLEANS                  *
#*****************************************/
PGA_TRUE=                   1
PGA_FALSE=                  0

#/*****************************************
#*                FLAGS                   *
#*****************************************/
PGA_FATAL=                 1
PGA_WARNING=               2

#/*****************************************
#*             MISC CONSTANT              *
#*****************************************/
PGA_TEMP1=                -1138
PGA_TEMP2 =               -4239
PGA_OLDPOP =              -6728
PGA_NEWPOP  =             -8376
PGA_UNINITIALIZED_INT   = -3827
PGA_UNINITIALIZED_DOUBLE= -968.3827


#/*****************************************
#*        DEBUG LEVELS                    *
#*****************************************/
PGA_DEBUG_ENTERED =       12
PGA_DEBUG_EXIT    =       13
PGA_DEBUG_MALLOC   =      80
PGA_DEBUG_PRINTVAR  =     82
PGA_DEBUG_SEND      =     22
PGA_DEBUG_RECV       =    23
PGA_DEBUG_MAXPGAPACKFUNCTIONS=	300
cdef enum:
   PGA_DEBUG_MAXFLAGS=1000

#/*****************************************
#*           DIRECTION                    *
#*****************************************/
PGA_MAXIMIZE=            1    
PGA_MINIMIZE=            2    
    
#/*****************************************
#*         STOPPING CRITERIA              *
#*****************************************/
PGA_STOP_MAXITER=        1    
PGA_STOP_NOCHANGE=       2    
PGA_STOP_TOOSIMILAR=     4    

#/*****************************************
#*            CROSSOVER                   *
#*****************************************/
PGA_CROSSOVER_ONEPT=     1    
PGA_CROSSOVER_TWOPT=     2    
PGA_CROSSOVER_UNIFORM=   3    

#/*****************************************
#*            SELECTION                   *
#*****************************************/
PGA_SELECT_PROPORTIONAL= 1    
PGA_SELECT_SUS=          2    
PGA_SELECT_TOURNAMENT=   3    
PGA_SELECT_PTOURNAMENT=  4    

#/*****************************************
#*            FITNESS                     *
#*****************************************/
PGA_FITNESS_RAW=         1    
PGA_FITNESS_NORMAL=      2    
PGA_FITNESS_RANKING=     3    

#/*****************************************
#*            FITNESS (MINIMIZATION)      *
#*****************************************/
PGA_FITNESSMIN_RECIPROCAL=  1 
PGA_FITNESSMIN_CMAX=        2 

#/*****************************************
#*               MUTATION                 *
#*****************************************/
PGA_MUTATION_CONSTANT=   1    
PGA_MUTATION_RANGE=      2    
PGA_MUTATION_UNIFORM=    3    
PGA_MUTATION_GAUSSIAN=   4    
PGA_MUTATION_PERMUTE=    5    
    
#/*****************************************
#*        POPULATION REPLACEMENT          *
#*****************************************/
PGA_POPREPL_BEST=         1   
PGA_POPREPL_RANDOM_NOREP= 2   
PGA_POPREPL_RANDOM_REP=   3   

#/****************************************
# *       REPORT OPTIONS                 *
# ****************************************/
PGA_REPORT_ONLINE=        1    
PGA_REPORT_OFFLINE=       2    
PGA_REPORT_HAMMING=       4    
PGA_REPORT_STRING=        8    
PGA_REPORT_WORST=         16   
PGA_REPORT_AVERAGE=       32   

#/*****************************************
#*            RANDOMIZER                  *
#*****************************************/
PGA_RINIT_PERCENT=        1  
PGA_RINIT_RANGE=          2  
PGA_IINIT_PERMUTE=        1  
PGA_IINIT_RANGE=          2  
PGA_CINIT_LOWER=          1  
PGA_CINIT_UPPER=          2  
PGA_CINIT_MIXED=          3  

##/*****************************************
#*         SET USER FUNCTION              *
#*****************************************/
PGA_USERFUNCTION_CREATESTRING=            1
PGA_USERFUNCTION_MUTATION=                2
PGA_USERFUNCTION_CROSSOVER=               3
PGA_USERFUNCTION_PRINTSTRING=             4
PGA_USERFUNCTION_COPYSTRING=              5
PGA_USERFUNCTION_DUPLICATE=               6
PGA_USERFUNCTION_INITSTRING=              7
PGA_USERFUNCTION_BUILDDATATYPE=           8
PGA_USERFUNCTION_STOPCOND=                9
PGA_USERFUNCTION_ENDOFGEN=                10
PGA_NUM_USERFUNCTIONS=                    10

#/*****************************************
#*           MPI SEND/RECV TAGS           *
#*****************************************/
PGA_COMM_STRINGTOEVAL=        1  
PGA_COMM_EVALOFSTRING=        2  
PGA_COMM_DONEWITHEVALS=       3  

cdef extern from "pgapack.h":
#    cdef struct PGAIndividual
#    cdef struct PGAAlgorithm

    ctypedef struct PGAIndividual:
      double evalfunc
      double fitness
      int    evaluptodate
      void   *chrom

    ctypedef struct PGAAlgorithm:
        int datatype
        int optdir
        int tw
        int fw
        int eb
        int PopSize
        int StringLen
        int StoppingRule
        int MaxIter
        int MaxNoChange
        int MaxSimilarity
        int NumReplace
        int PopReplace
        int iter
        int ItersOfSame
        int PercentSame
        int NoDuplicates
        int CrossoverType
        int SelectType
        int SelectIndex
        int FitnessType
        int FitnessMinType
        int MutateOnlyNoCross
        int MutationType
        int MutateIntegerValue
        int MutateBoundedFlag
        double MutateRealValue
        double MutationProb
        double CrossoverProb
        double UniformCrossProb
        double PTournamentProb
        double FitnessRankMax
        double FitnessCmaxValue
        double restartAlleleProb
        int restart
        int restartFreq
        int *selected
        int *sorted
        PGAIndividual *oldpop
        PGAIndividual *newpop

    ctypedef struct PGAContext
    cdef struct FILE

    ctypedef void * MPI_Datatype
    ctypedef void * MPI_Comm
    ctypedef void * MPI_Aint
    cdef struct MPI_Status

    ctypedef struct PGACOperations:
        void         (*CreateString)(PGAContext *, int, int, int)
        int          (*Mutation)(PGAContext *, int, int, double)
        void         (*Crossover)(PGAContext *, int, int, int, int, int, int)
        void         (*PrintString)(PGAContext *, FILE *, int, int)
        void         (*CopyString)(PGAContext *, int, int, int, int)
        int          (*Duplicate)(PGAContext *, int, int, int, int)
        void         (*InitString)(PGAContext *, int, int)
        MPI_Datatype (*BuildDatatype)(PGAContext *, int, int)
        int          (*StopCond)(PGAContext *)
        void         (*EndOfGen)(PGAContext *)

    ctypedef struct PGAFortranOperations:
        int          (*Mutation)(void *, void *, void *, void *)
        void         (*Crossover)(void *, void *, void *, void *, void *, void *, void *)
        void         (*PrintString)(void *, void *, void *, void *)
        void         (*CopyString)(void *, void *, void *, void *, void *)
        int          (*Duplicate)(void *, void *, void *, void *, void *)
        void         (*InitString)(void *, void *, void *)
        int          (*StopCond)(void *)
        void         (*EndOfGen)(void *)

    ctypedef struct PGAParallel:
        int      MPIAlreadyInit
        int      NumIslands
        int      NumDemes
        MPI_Comm DefaultComm
        int      MPIStubLibrary

    ctypedef int time_t

    ctypedef struct PGAReport:
         int    PrintFreq
         int    PrintOptions
         double Offline
         double Online
         double Average
         double Best
         time_t starttime

    ctypedef struct PGASystem:
        int    UserFortran
        int    SetUpCalled
        int    PGAMaxInt
        int    PGAMinInt
        double PGAMaxDouble
        double PGAMinDouble

    ctypedef struct PGADebug:
         int PGADebugFlags[PGA_DEBUG_MAXFLAGS]

    ctypedef struct PGAInitialize:
        int    RandomInit
        double BinaryProbability
        int    RealType
        int    IntegerType
        int    CharacterType
        int    *IntegerMin
        int    *IntegerMax
        double *RealMin
        double *RealMax
        int    RandomSeed

    ctypedef struct PGAScratch:
        int    *intscratch
        double *dblscratch

#    cdef extern struct PGAContext

    #/*****************************************
    #*          binary.c
    #*****************************************/

    cdef  void PGASetBinaryAllele ( PGAContext *ctx, int p, int pop, int i, int val )
    cdef  int PGAGetBinaryAllele ( PGAContext *ctx, int p, int pop, int i )
    cdef  void PGASetBinaryInitProb ( PGAContext *ctx, double probability )
    cdef  double PGAGetBinaryInitProb (PGAContext *ctx)
    cdef  void PGABinaryCreateString(PGAContext *ctx, int p, int pop, int initflag)
    cdef  int PGABinaryMutation( PGAContext *ctx, int p, int pop, double mr )
    cdef  void PGABinaryOneptCrossover(PGAContext *ctx, int p1, int p2, int pop1, int c1,
                         int c2, int pop2)
    cdef  void PGABinaryTwoptCrossover(PGAContext *ctx, int p1, int p2, int pop1, int c1,
                         int c2, int pop2)
    cdef  void PGABinaryUniformCrossover(PGAContext *ctx, int p1, int p2, int pop1,
                           int c1, int c2, int pop2)
    cdef  void PGABinaryPrintString( PGAContext *ctx, FILE *fp, int p, int pop )
    cdef  void PGABinaryCopyString (PGAContext *ctx, int p1, int pop1, int p2, int pop2)
    cdef  int PGABinaryDuplicate( PGAContext *ctx, int p1, int pop1, int p2, int pop2)
    cdef  void PGABinaryInitString(PGAContext *ctx, int p, int pop)
    cdef  MPI_Datatype PGABinaryBuildDatatype(PGAContext *ctx, int p, int pop)
    cdef  int PGABinaryHammingDistance ( PGAContext *ctx, PGABinary *s1, PGABinary *s2 )
    cdef  void PGABinaryPrint( PGAContext *ctx, FILE *fp, PGABinary *chrom, int nb )

    #/*****************************************
    #*  char.c
    #*****************************************/

    cdef  void PGASetCharacterAllele (PGAContext *ctx, int p, int pop, int i, char value)
    cdef  char PGAGetCharacterAllele (PGAContext *ctx, int p, int pop, int i)
    cdef  void PGASetCharacterInitType(PGAContext *ctx, int value)
    cdef  void PGACharacterCreateString (PGAContext *ctx, int p, int pop, int InitFlag)
    cdef  int PGACharacterMutation( PGAContext *ctx, int p, int pop, double mr )
    cdef  void PGACharacterOneptCrossover(PGAContext *ctx, int p1, int p2, int pop1,
                            int c1, int c2, int pop2)
    cdef  void PGACharacterTwoptCrossover( PGAContext *ctx, int p1, int p2, int pop1,
                          int c1, int c2, int pop2)
    cdef  void PGACharacterUniformCrossover(PGAContext *ctx, int p1, int p2, int pop1,
                            int c1, int c2, int pop2)
    cdef  void PGACharacterPrintString ( PGAContext *ctx, FILE *fp, int p, int pop)
    cdef  void PGACharacterCopyString (PGAContext *ctx, int p1, int pop1, int p2,
                         int pop2)
    cdef  int PGACharacterDuplicate( PGAContext *ctx, int p1, int pop1, int p2, int pop2)
    cdef  void PGACharacterInitString(PGAContext *ctx, int p, int pop)
    cdef  MPI_Datatype PGACharacterBuildDatatype(PGAContext *ctx, int p, int pop)

    #/*****************************************
    #*  cmdline.c
    #*****************************************/

    cdef  void PGAReadCmdLine( PGAContext *ctx, int *argc, char **argv )
    cdef  void PGAParseDebugArg(PGAContext *ctx, char *st)
    cdef  void PGAStripArgs(char **argv, int *argc, int *c, int num)

    #/*****************************************
    #*  create.c
    #*****************************************/

    cdef  PGAContext *PGACreate ( int *argc, char **argv,
                    int datatype, int len, int maxormin )
    cdef  void PGASetUp ( PGAContext *ctx )
    cdef  void PGASetRandomInitFlag(PGAContext *ctx, int RandomBoolean)
    cdef  int PGAGetRandomInitFlag (PGAContext *ctx)
    cdef  void PGACreatePop (PGAContext *ctx, int pop)
    cdef  void PGACreateIndividual (PGAContext *ctx, int p, int pop, int initflag)

    #/*****************************************
    #*  cross.c
    #*****************************************/

    cdef  void PGACrossover ( PGAContext *ctx, int p1, int p2, int pop1,
                int c1, int c2, int pop2 )
    cdef  int PGAGetCrossoverType (PGAContext *ctx)
    cdef  double PGAGetCrossoverProb (PGAContext *ctx)
    cdef  double PGAGetUniformCrossoverProb (PGAContext *ctx)
    cdef  void PGASetCrossoverType (PGAContext *ctx, int crossover_type)
    cdef  void PGASetCrossoverProb( PGAContext *ctx, double crossover_prob)
    cdef  void PGASetUniformCrossoverProb( PGAContext *ctx, double uniform_cross_prob)

    #/*****************************************
    #*  debug.c
    #*****************************************/

    cdef  void PGASortFuncNameIndex(PGAContext *ctx)
    #if OPTIMIZE==0:
    #    cdef  void PGADebugPrint( PGAContext *ctx, int level, char *funcname,
    #           char *msg, int datatype, void *data )
           
#    cdef  void PGASetDebugLevel(PGAContext *ctx, int level)
#    cdef  void PGAClearDebugLevel(PGAContext *ctx, int level)
#    cdef  void PGASetDebugLevelByName(PGAContext *ctx, char *funcname)
#    cdef  void PGAClearDebugLevelByName(PGAContext *ctx, char *funcname)
#    cdef  int PGAGetDebugLevelOfName(PGAContext *ctx, char *funcname)
#    cdef  int PGAGetDebugFlag(PGAContext *ctx, char *funcname)
#    cdef  void PGASetDebugFlag11(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag20(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag21(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag30(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag32(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag34(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag36(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag40(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag42(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag44(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag46(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag48(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag50(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag52(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag54(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag56(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag58(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag60(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag62(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag64(PGAContext *ctx, int Flag)
#    cdef  void PGASetDebugFlag66(PGAContext *ctx, int Flag)
#    cdef  void PGAPrintDebugOptions(PGAContext *ctx)

    #/*****************************************
    #*  duplcate.c
    #*****************************************/

    cdef  int PGADuplicate(PGAContext *ctx, int p, int pop1, int pop2, int n)
    cdef  void PGAChange( PGAContext *ctx, int p, int pop )
    cdef  void PGASetNoDuplicatesFlag( PGAContext *ctx, int no_dup)
    cdef  int PGAGetNoDuplicatesFlag (PGAContext *ctx)

    #/*****************************************
    #*  evaluate.c
    #*****************************************/

    cdef  void PGASetEvaluation ( PGAContext *ctx, int p, int pop, double val )
    cdef  double PGAGetEvaluation ( PGAContext *ctx, int p, int pop )
    cdef  void PGASetEvaluationUpToDateFlag ( PGAContext *ctx, int p, int pop,
                               int status )
    cdef  int PGAGetEvaluationUpToDateFlag ( PGAContext *ctx, int p, int pop )
    cdef  double PGAGetRealFromBinary(PGAContext *ctx, int p, int pop, int start,
                        int end, double lower, double upper)
    cdef  double PGAGetRealFromGrayCode(PGAContext *ctx, int p, int pop, int start,
                              int end, double lower, double upper)
    cdef  void PGAEncodeRealAsBinary(PGAContext *ctx, int p, int pop, int start,
                           int end, double low, double high, double val)
    cdef  void PGAEncodeRealAsGrayCode(PGAContext *ctx, int p, int pop, int start,
                          int end, double low, double high, double val)
    cdef  int PGAGetIntegerFromBinary(PGAContext *ctx, int p, int pop, int start,
                             int end)
    cdef  int PGAGetIntegerFromGrayCode(PGAContext *ctx, int p, int pop, int start,
                               int end)
    cdef  void PGAEncodeIntegerAsBinary(PGAContext *ctx, int p, int pop, int start,
                          int end, int val)
    cdef  void PGAEncodeIntegerAsGrayCode(PGAContext *ctx, int p, int pop, int start,
                            int end, int val)
    cdef  double PGAMapIntegerToReal (PGAContext *ctx, int v, int a, int b, double l,
                        double u)
    cdef  int PGAMapRealToInteger(PGAContext *ctx, double r, double l, double u, int a,
                    int b)

    #/*****************************************
    #*  fitness.c
    #*****************************************/

    cdef  void PGAFitness ( PGAContext *ctx, int popindex )
    cdef  int PGARank( PGAContext *ctx, int p, int *order, int n )
    cdef  double PGAGetFitness ( PGAContext *ctx, int p, int pop )
    cdef  int PGAGetFitnessType (PGAContext *ctx)
    cdef  int PGAGetFitnessMinType (PGAContext *ctx)
    cdef  double PGAGetMaxFitnessRank (PGAContext *ctx)
    cdef  void PGASetFitnessType( PGAContext *ctx, int fitness_type)
    cdef  void PGASetFitnessMinType( PGAContext *ctx, int fitness_type)
    cdef  void PGASetMaxFitnessRank( PGAContext *ctx, double fitness_rank_max)
    cdef  void PGAFitnessLinearNormal ( PGAContext *ctx, PGAIndividual *pop )
    cdef  void PGAFitnessLinearRank ( PGAContext *ctx, PGAIndividual *pop )
    cdef  void PGAFitnessMinReciprocal ( PGAContext *ctx, PGAIndividual *pop )
    cdef  void PGAFitnessMinCmax ( PGAContext *ctx, PGAIndividual *pop )
    cdef  void PGASetFitnessCmaxValue( PGAContext *ctx, double val)
    cdef  double PGAGetFitnessCmaxValue (PGAContext *ctx)

    #/*****************************************
    #*  hamming.c
    #*****************************************/

    cdef  double PGAHammingDistance( PGAContext *ctx, int popindex)

    #/*****************************************
    #*  heap.c
    #*****************************************/

    cdef  void PGADblHeapSort ( PGAContext *ctx, double *a, int *idx, int n )
    cdef  void PGAIntHeapSort ( PGAContext *ctx, int *a, int *idx, int n )

    #/*****************************************
    #*  integer.c
    #*****************************************/

    cdef  void PGASetIntegerAllele (PGAContext *ctx, int p, int pop, int i, int value)
    cdef  int PGAGetIntegerAllele (PGAContext *ctx, int p, int pop, int i)
    cdef  void PGASetIntegerInitPermute ( PGAContext *ctx, int min, int max)
    cdef  void PGASetIntegerInitRange (PGAContext *ctx, int *min, int *max)
    cdef  int PGAGetIntegerInitType (PGAContext *ctx)
    cdef  int PGAGetMinIntegerInitValue (PGAContext *ctx, int i)
    cdef  int PGAGetMaxIntegerInitValue (PGAContext *ctx, int i)
    cdef  void PGAIntegerCreateString (PGAContext *ctx, int p, int pop, int InitFlag)
    cdef  int PGAIntegerMutation( PGAContext *ctx, int p, int pop, double mr )
    cdef  void PGAIntegerOneptCrossover(PGAContext *ctx, int p1, int p2, int pop1,
                          int c1, int c2, int pop2)
    cdef  void PGAIntegerTwoptCrossover( PGAContext *ctx, int p1, int p2, int pop1,
                          int c1, int c2, int pop2)
    cdef  void PGAIntegerUniformCrossover(PGAContext *ctx, int p1, int p2, int pop1,
                            int c1, int c2, int pop2)
    cdef  void PGAIntegerPrintString ( PGAContext *ctx, FILE *fp, int p, int pop)
    cdef  void PGAIntegerCopyString (PGAContext *ctx, int p1, int pop1, int p2, int pop2)
    cdef  int PGAIntegerDuplicate( PGAContext *ctx, int p1, int pop1, int p2, int pop2)
    cdef  void PGAIntegerInitString(PGAContext *ctx, int p, int pop)
    cdef  MPI_Datatype PGAIntegerBuildDatatype(PGAContext *ctx, int p, int pop)

    #/*****************************************
    #*  mpi_stub.c
    #*****************************************/


    #/*****************************************
    #*  mutation.c
    #*****************************************/

    cdef  int PGAMutate(PGAContext *ctx, int p, int pop)
    cdef  void PGASetMutationType( PGAContext *ctx, int mutation_type)
    cdef  int PGAGetMutationType (PGAContext *ctx)
    cdef  void PGASetMutationRealValue( PGAContext *ctx, double val)
    cdef  double PGAGetMutationRealValue (PGAContext *ctx)
    cdef  void PGASetMutationIntegerValue( PGAContext *ctx, int val)
    cdef  int PGAGetMutationIntegerValue (PGAContext *ctx)
    cdef  void PGASetMutationBoundedFlag(PGAContext *ctx, int val)
    cdef  int PGAGetMutationBoundedFlag(PGAContext *ctx)
    cdef  void PGASetMutationProb(PGAContext *ctx, double mutation_prob)
    cdef  double PGAGetMutationProb (PGAContext *ctx)

    #/*****************************************
    #*  parallel.c
    #*****************************************/

    cdef  void PGARunGM(PGAContext *ctx, double (*f)(PGAContext *, int, int), MPI_Comm comm)
    cdef  void PGAEvaluateSeq(PGAContext *ctx, int pop, double (*f)(PGAContext *, int, int))
    cdef  void PGAEvaluateCoop(PGAContext *ctx, int pop, double (*f)(PGAContext *, int, int), MPI_Comm comm)
    cdef  void PGAEvaluateMS(PGAContext *ctx, int pop, double (*f)(PGAContext *c, int p, int pop), MPI_Comm comm)
    cdef  void PGAEvaluateSlave(PGAContext *ctx, int pop, double (*f)(PGAContext *, int, int), MPI_Comm comm)
    cdef  void PGAEvaluate(PGAContext *ctx, int pop, double (*f)(PGAContext *, int, int), MPI_Comm comm)
    cdef  MPI_Datatype PGABuildDatatype(PGAContext *ctx, int p, int pop)
    cdef  void PGASendIndividual(PGAContext *ctx, int p, int pop, int dest, int tag, MPI_Comm comm)
    cdef  void PGAReceiveIndividual(PGAContext *ctx, int p, int pop, int source, int tag, MPI_Comm comm, MPI_Status *status)
    cdef  void PGASendReceiveIndividual(PGAContext *ctx, int send_p, int send_pop, int dest, int send_tag, int recv_p, int recv_pop, int source, int recv_tag, MPI_Comm comm, MPI_Status *status)
    cdef  void PGARunIM(PGAContext *ctx, double (*f)(PGAContext *c, int p, int pop), MPI_Comm tcomm)
    cdef  void PGARunNM(PGAContext *ctx, double (*f)(PGAContext *c, int p, int pop), MPI_Comm tcomm)
    cdef  int PGAGetRank (PGAContext *ctx, MPI_Comm comm)
    cdef  int PGAGetNumProcs (PGAContext *ctx, MPI_Comm comm)
    cdef  void PGASetNumIslands( PGAContext *ctx, int n)
    cdef  int PGAGetNumIslands (PGAContext *ctx)
    cdef  void PGASetNumDemes( PGAContext *ctx, int numdemes)
    cdef  int PGAGetNumDemes (PGAContext *ctx)
    cdef  void PGASetCommunicator( PGAContext *ctx, MPI_Comm comm)
    cdef  MPI_Comm PGAGetCommunicator( PGAContext *ctx)

    #/*****************************************
    #*  pga.c
    #*****************************************/

    cdef  void PGARun(PGAContext *ctx, double (*evaluate)(PGAContext *c, int p, int pop))
    cdef  void PGARunMutationAndCrossover (PGAContext *ctx, int oldpop, int newpop)
    cdef  void PGARunMutationOrCrossover ( PGAContext *ctx, int oldpop, int newpop )
    cdef  void PGAUpdateGeneration(PGAContext *ctx, MPI_Comm comm)
    cdef  int PGAGetDataType (PGAContext *ctx)
    cdef  int PGAGetOptDirFlag (PGAContext *ctx)
    cdef  int PGAGetStringLength (PGAContext *ctx)
    cdef  int PGAGetVariableStringLength (PGAContext *ctx, int p, int pop)
    cdef  int PGAGetGAIterValue (PGAContext *ctx)
    cdef  void PGASetMutationOrCrossoverFlag( PGAContext *ctx, int flag)
    cdef  void PGASetMutationAndCrossoverFlag( PGAContext *ctx, int flag)
    cdef  int PGAGetMutationOrCrossoverFlag (PGAContext *ctx)
    cdef  int PGAGetMutationAndCrossoverFlag (PGAContext *ctx)

    #/*****************************************
    #*  pop.c
    #*****************************************/

    cdef  void PGASortPop ( PGAContext *ctx, int pop )
    cdef  int PGAGetPopSize (PGAContext *ctx)
    cdef  int PGAGetNumReplaceValue (PGAContext *ctx)
    cdef  int PGAGetPopReplaceType (PGAContext *ctx)
    cdef  int PGAGetSortedPopIndex ( PGAContext *ctx, int n )
    cdef  void PGASetPopSize (PGAContext *ctx, int popsize)
    cdef  void PGASetNumReplaceValue( PGAContext *ctx, int pop_replace)
    cdef  void PGASetPopReplaceType( PGAContext *ctx, int pop_replace)

    #/*****************************************
    #*  random.c
    #*****************************************/

    cdef  int PGARandomFlip ( PGAContext *ctx, double p )
    cdef  int PGARandomInterval( PGAContext *ctx, int start, int end)
    cdef  double PGARandom01( PGAContext *ctx, int newseed )
    cdef  double PGARandomUniform( PGAContext *ctx, double start, double end)
    cdef  double PGARandomGaussian( PGAContext *ctx, double mean, double sigma)
    cdef  int PGAGetRandomSeed(PGAContext *ctx)
    cdef  void PGASetRandomSeed(PGAContext *ctx, int seed)

    #/*****************************************
    #*  real.c
    #*****************************************/

    cdef  void PGASetRealAllele (PGAContext *ctx, int p, int pop, int i, double value)
    cdef  double PGAGetRealAllele (PGAContext *ctx, int p, int pop, int i)
    cdef  void PGASetRealInitPercent ( PGAContext *ctx, double *median, double *percent)
    cdef  void PGASetRealInitRange (PGAContext *ctx, double *min, double *max)
    cdef  double PGAGetMinRealInitValue (PGAContext *ctx, int i)
    cdef  double PGAGetMaxRealInitValue (PGAContext *ctx, int i)
    cdef  int PGAGetRealInitType (PGAContext *ctx)
    cdef  void PGARealCreateString (PGAContext *ctx, int p, int pop, int initflag)
    cdef  int PGARealMutation( PGAContext *ctx, int p, int pop, double mr )
    cdef  void PGARealOneptCrossover( PGAContext *ctx, int p1, int p2, int pop1,
                       int c1, int c2, int pop2)
    cdef  void PGARealTwoptCrossover( PGAContext *ctx, int p1, int p2, int pop1,
                       int c1, int c2, int pop2)
    cdef  void PGARealUniformCrossover( PGAContext *ctx, int p1, int p2, int pop1,
                         int c1, int c2, int pop2)
    cdef  void PGARealPrintString (PGAContext *ctx, FILE *fp, int p, int pop)
    cdef  void PGARealCopyString ( PGAContext *ctx, int p1, int pop1, int p2, int pop2)
    cdef  int PGARealDuplicate( PGAContext *ctx, int p1, int pop1, int p2, int pop2)
    cdef  void PGARealInitString ( PGAContext *ctx, int p, int pop)
    cdef  MPI_Datatype PGARealBuildDatatype(PGAContext *ctx, int p, int pop)

    #/*****************************************
    #*  report.c
    #*****************************************/

    cdef  void PGAPrintReport(PGAContext *ctx, FILE *fp, int pop)
    cdef  void PGASetPrintOptions (PGAContext *ctx, int option)
    cdef  void PGASetPrintFrequencyValue( PGAContext *ctx, int print_freq)
    cdef  int PGAGetPrintFrequencyValue (PGAContext *ctx)
    cdef  void PGAPrintPopulation ( PGAContext *ctx, FILE *fp, int pop )
    cdef  void PGAPrintIndividual ( PGAContext *ctx, FILE *fp, int p, int pop )
    cdef  void PGAPrintString ( PGAContext *ctx, FILE *file, int p, int pop )
    cdef  void PGAPrintContextVariable ( PGAContext *ctx, FILE *fp )

    #/*****************************************
    #*  restart.c
    #*****************************************/

    cdef  void PGARestart(PGAContext *ctx, int source_pop, int dest_pop)
    cdef  void PGASetRestartFlag(PGAContext *ctx, int val)
    cdef  int PGAGetRestartFlag(PGAContext *ctx)
    cdef  void PGASetRestartFrequencyValue(PGAContext *ctx, int numiter)
    cdef  int PGAGetRestartFrequencyValue(PGAContext *ctx)
    cdef  void PGASetRestartAlleleChangeProb(PGAContext *ctx, double prob)
    cdef  double PGAGetRestartAlleleChangeProb(PGAContext *ctx)

    #/*****************************************
    #*  select.c
    #*****************************************/

    cdef  void PGASelect( PGAContext *ctx, int popix )
    cdef  int PGASelectNextIndex ( PGAContext *ctx )
    cdef  void PGASetSelectType( PGAContext *ctx, int select_type)
    cdef  int PGAGetSelectType (PGAContext *ctx)
    cdef  void PGASetPTournamentProb(PGAContext *ctx, double ptournament_prob)
    cdef  double PGAGetPTournamentProb(PGAContext *ctx)
    cdef  int PGASelectProportional(PGAContext *ctx, PGAIndividual *pop)
    cdef  void PGASelectSUS( PGAContext *ctx, PGAIndividual *pop )
    cdef  int PGASelectTournament( PGAContext *ctx, PGAIndividual *pop )
    cdef  int PGASelectPTournament( PGAContext *ctx, PGAIndividual *pop )

    #/*****************************************
    #*  stop.c
    #*****************************************/

    cdef  int PGADone(PGAContext *ctx, MPI_Comm comm)
    cdef  int PGACheckStoppingConditions( PGAContext *ctx)
    cdef  void PGASetStoppingRuleType (PGAContext *ctx, int stoprule)
    cdef  int PGAGetStoppingRuleType (PGAContext *ctx)
    cdef  void PGASetMaxGAIterValue(PGAContext *ctx, int maxiter)
    cdef  int PGAGetMaxGAIterValue (PGAContext *ctx)
    cdef  void PGASetMaxNoChangeValue(PGAContext *ctx, int max_no_change)
    cdef  void PGASetMaxSimilarityValue(PGAContext *ctx, int max_similarity)

    #/*****************************************
    #*  system.c
    #*****************************************/

    cdef  void PGAError( PGAContext *ctx, char *msg,
           int level, int datatype, void *data )
    cdef  void PGADestroy (PGAContext *ctx)
    cdef  int PGAGetMaxMachineIntValue (PGAContext *ctx)
    cdef  int PGAGetMinMachineIntValue (PGAContext *ctx)
    cdef  double PGAGetMaxMachineDoubleValue (PGAContext *ctx)
    cdef  double PGAGetMinMachineDoubleValue (PGAContext *ctx)
    cdef  void PGAUsage( PGAContext *ctx )
    cdef  void PGAPrintVersionNumber( PGAContext *ctx )

    #/*****************************************
    #*  user.c
    #*****************************************/

    cdef  void PGASetUserFunction(PGAContext *ctx, int constant, void *f)

    #/*****************************************
    #*  utility.c
    #*****************************************/

    cdef  double PGAMean ( PGAContext *ctx, double *a, int n)
    cdef  double PGAStddev ( PGAContext *ctx, double *a, int n, double mean)
    cdef  int PGARound(PGAContext *ctx, double x)
    cdef  void PGACopyIndividual( PGAContext *ctx, int p1, int pop1, int p2, int pop2)
    cdef  int PGACheckSum(PGAContext *ctx, int p, int pop)
    cdef  int PGAGetWorstIndex(PGAContext *ctx, int pop)
    cdef  int PGAGetBestIndex(PGAContext *ctx, int pop)
    cdef  PGAIndividual *PGAGetIndividual ( PGAContext *ctx, int p, int pop)
    cdef  void PGAUpdateAverage(PGAContext *ctx, int pop)
    cdef  void PGAUpdateOnline(PGAContext *ctx, int pop)
    cdef  void PGAUpdateOffline(PGAContext *ctx, int pop)
    cdef  int PGAComputeSimilarity(PGAContext *ctx, PGAIndividual *pop)

ctypedef struct PGAContext:
     PGAAlgorithm           ga
     PGACOperations         cops
     PGAFortranOperations   fops
     PGAParallel            par
     PGAReport              rep
     PGASystem              sys
     PGADebug               debug
     PGAInitialize          init
     PGAScratch             scratch
     void                   *extra_args

cdef extern from "arrayobject.h":

    cdef void import_array()

    cdef enum PyArray_TYPES:
                 PyArray_CHAR
                 PyArray_UBYTE
                 PyArray_SBYTE
                 PyArray_SHORT
                 PyArray_USHORT
                 PyArray_INT
                 PyArray_UINT
                 PyArray_LONG
                 PyArray_FLOAT
                 PyArray_DOUBLE
                 PyArray_CFLOAT
                 PyArray_CDOUBLE
                 PyArray_OBJECT
                 PyArray_NTYPES
                 PyArray_NOTYPE

    struct PyArray_Descr:
        int type_num, elsize
        char type

    ctypedef class numpy.ndarray [object PyArrayObject]:
        cdef char *data
        cdef int nd
        cdef int *dimensions, *strides
        cdef object base
        cdef PyArray_Descr *descr
        cdef int flags

    cdef object PyArray_FromObject(object x, char typecode, int dmin, int dmax)

cdef double Evaluate((PGAContext *)ctx, int p, int pop):
     cdef int naxis
     cdef int *naxes
     cdef double a
     N = PGAGetStringLength(ctx)
     NA = 1
     dtype = PGAGetDataType(ctx)
     kind = [PyArray_INT, PyArray_INT, PyArray_DOUBLE, PyArray_CHAR][dtype-1]
     naxes = <int *>malloc(NA*sizeof(int))
     alleles = []
     for i in range(N):
          if dtype == 1: val = PGAGetBinaryAllele(ctx,p,pop,i)
          elif dtype == 2: val = PGAGetIntegerAllele(ctx,p,pop,i)
          elif dtype == 3: val = PGAGetRealAllele(ctx,p,pop,i)
          elif dtype == 4: val = PGAGetCharacterAllele(ctx,p,pop,i)
          alleles.append(val)
     func = <object> ctx.extra_args
     x = PyArray_FromObject(alleles,kind,1,1)
     a = func(x)
     return a

cdef class Context:
    cdef PGAContext *ctx
    cdef int *argc
    cdef char **argv
    cdef int datatype
    cdef int len
    cdef int maxormin
    cdef int setup
    cdef object func

    def __new__(self,pdatatype,plen,pmaxormin,in_argv=[]):
        import_array()
        cdef int in_argc
        cdef char *tmp
        in_argc = len(in_argv)
        self.datatype = pdatatype
        self.len = plen
        self.maxormin = pmaxormin
        la = 0
        self.argc = &in_argc
        self.argv = <char **>malloc(la*sizeof(char))
        self.ctx = <PGAContext *>PGACreate(self.argc,self.argv,self.datatype,self.len,self.maxormin)
        self.setup = 0 
        self.func = None

    def __dealloc__(self):
        PGADestroy(self.ctx)
        free(self.argv)

    def SetUp(self):
        PGASetUp(self.ctx)

    def Run(self,evalfunc):
        self.func = evalfunc
        self.ctx.extra_args = <void*>self.func
        Py_INCREF(self.func)
        PGARun(self.ctx, Evaluate)

    def Set(self,item,*params,**keyw):
        cdef ndarray a
        cdef ndarray b
        if item == "BinaryAllele":
           PGASetBinaryAllele ( self.ctx,params[0], params[1],params[2],params[3] )
        elif item == "BinaryInitProb":
           PGASetBinaryInitProb ( self.ctx,  params[0] )
        elif item == "CharacterAllele":
           PGASetCharacterAllele (self.ctx, params[0], params[1], params[2],  params[3])
        elif item == "CharacterInitType":
           PGASetCharacterInitType(self.ctx, params[0])
#        elif item == "Up":
#           PGASetUp ( self.ctx )
        elif item == "RandomInitFlag":
           PGASetRandomInitFlag(self.ctx, params[0])
        elif item == "CrossoverType":
           PGASetCrossoverType (self.ctx, params[0])
        elif item == "CrossoverProb":
           PGASetCrossoverProb( self.ctx,  params[0])
        elif item == "UniformCrossoverProb":
           PGASetUniformCrossoverProb( self.ctx,  params[0])
#        elif item == "DebugLevel":
#           level= value
#           PGASetDebugLevel(self.ctx, level)
#        elif item == "DebugLevelByName":
#            *funcname= value
#           PGASetDebugLevelByName(self.ctx,  *funcname)
#        elif item == "DebugFlag11":
#           Flag= value
#           PGASetDebugFlag11(self.ctx, Flag)
#        elif item == "DebugFlag20":
#           Flag= value
#           PGASetDebugFlag20(self.ctx, Flag)
#        elif item == "DebugFlag21":
#           Flag= value
#           PGASetDebugFlag21(self.ctx, Flag)
#        elif item == "DebugFlag30":
#           Flag= value
#           PGASetDebugFlag30(self.ctx, Flag)
#        elif item == "DebugFlag32":
#           Flag= value
#           PGASetDebugFlag32(self.ctx, Flag)
#        elif item == "DebugFlag34":
#           Flag= value
#           PGASetDebugFlag34(self.ctx, Flag)
#        elif item == "DebugFlag36":
#           Flag= value
#           PGASetDebugFlag36(self.ctx, Flag)
#        elif item == "DebugFlag40":
#           Flag= value
#           PGASetDebugFlag40(self.ctx, Flag)
#        elif item == "DebugFlag42":
#           Flag= value
#           PGASetDebugFlag42(self.ctx, Flag)
#        elif item == "DebugFlag44":
#           Flag= value
#           PGASetDebugFlag44(self.ctx, Flag)
#        elif item == "DebugFlag46":
#           Flag= value
#           PGASetDebugFlag46(self.ctx, Flag)
#        elif item == "DebugFlag48":
#           Flag= value
#           PGASetDebugFlag48(self.ctx, Flag)
#        elif item == "DebugFlag50":
#           Flag= value
#           PGASetDebugFlag50(self.ctx, Flag)
#        elif item == "DebugFlag52":
#           Flag= value
#           PGASetDebugFlag52(self.ctx, Flag)
#        elif item == "DebugFlag54":
#           Flag= value
#           PGASetDebugFlag54(self.ctx, Flag)
#        elif item == "DebugFlag56":
#           Flag= value
#           PGASetDebugFlag56(self.ctx, Flag)
#        elif item == "DebugFlag58":
#           Flag= value
#           PGASetDebugFlag58(self.ctx, Flag)
#        elif item == "DebugFlag60":
#           Flag= value
#           PGASetDebugFlag60(self.ctx, Flag)
#        elif item == "DebugFlag62":
#           Flag= value
#           PGASetDebugFlag62(self.ctx, Flag)
#        elif item == "DebugFlag64":
#           Flag= value
#           PGASetDebugFlag64(self.ctx, Flag)
#        elif item == "DebugFlag66":
#           Flag= value
#           PGASetDebugFlag66(self.ctx, Flag)
        elif item == "NoDuplicatesFlag":
           PGASetNoDuplicatesFlag( self.ctx, params[0])
        elif item == "Evaluation":
           PGASetEvaluation ( self.ctx, params[0], params[1],  params[1] )
        elif item == "EvaluationUpToDateFlag":
           PGASetEvaluationUpToDateFlag ( self.ctx, params[0], params[1], params[2] )
        elif item == "FitnessType":
           PGASetFitnessType( self.ctx, params[0])
        elif item == "FitnessMinType":
           PGASetFitnessMinType( self.ctx, params[0])
        elif item == "MaxFitnessRank":
           PGASetMaxFitnessRank( self.ctx,  params[0])
        elif item == "FitnessCmaxValue":
           PGASetFitnessCmaxValue( self.ctx,  params[0])
        elif item == "IntegerAllele":
           PGASetIntegerAllele (self.ctx, params[0], params[1], params[2], params[3])
        elif item == "IntegerInitPermute":
           PGASetIntegerInitPermute ( self.ctx, params[0], params[1])
        elif item == "IntegerInitRange":
           a = <ndarray>params[0]
           b = <ndarray>params[1]
           PGASetIntegerInitRange (self.ctx, <int *>a.data, <int *>a.data)
        elif item == "MutationType":
           PGASetMutationType( self.ctx, params[0])
        elif item == "MutationRealValue":
           PGASetMutationRealValue( self.ctx,  params[0])
        elif item == "MutationIntegerValue":
           PGASetMutationIntegerValue( self.ctx, params[0])
        elif item == "MutationBoundedFlag":
           PGASetMutationBoundedFlag(self.ctx, params[0])
        elif item == "MutationProb":
           PGASetMutationProb(self.ctx,  params[0])
        elif item == "NumIslands":
           PGASetNumIslands( self.ctx, params[0])
        elif item == "NumDemes":
           PGASetNumDemes( self.ctx, params[0])
#        elif item == "Communicator":
#           comm= value
#           PGASetCommunicator( self.ctx, comm)
        elif item == "MutationOrCrossoverFlag":
           PGASetMutationOrCrossoverFlag( self.ctx, params[0])
        elif item == "MutationAndCrossoverFlag":
           PGASetMutationAndCrossoverFlag( self.ctx, params[0])
        elif item == "PopSize":
           PGASetPopSize (self.ctx, params[0])
        elif item == "NumReplaceValue":
           PGASetNumReplaceValue( self.ctx, params[0])
        elif item == "PopReplaceType":
           PGASetPopReplaceType( self.ctx, params[0])
        elif item == "RandomSeed":
           PGASetRandomSeed(self.ctx, params[0])
        elif item == "RealAllele":
           PGASetRealAllele (self.ctx, params[0], params[0], params[0],  params[0])
        elif item == "RealInitPercent":
           a = <ndarray>params[0]
           b = <ndarray>params[1]
           PGASetRealInitPercent ( self.ctx,  <double *>a.data, <double *>b.data)
        elif item == "RealInitRange":
           a = <ndarray>params[0]
           b = <ndarray>params[1]
           PGASetRealInitRange (self.ctx, <double *>a.data, <double *>b.data)
        elif item == "PrintOptions":
           PGASetPrintOptions (self.ctx, params[0])
        elif item == "PrintFrequencyValue":
           PGASetPrintFrequencyValue( self.ctx, params[0])
        elif item == "RestartFlag":
           PGASetRestartFlag(self.ctx, params[0])
        elif item == "RestartFrequencyValue":
           PGASetRestartFrequencyValue(self.ctx, params[0])
        elif item == "RestartAlleleChangeProb":
           PGASetRestartAlleleChangeProb(self.ctx,  params[0])
        elif item == "SelectType":
           PGASetSelectType( self.ctx, params[0])
        elif item == "PTournamentProb":
           PGASetPTournamentProb(self.ctx,  params[0])
        elif item == "StoppingRuleType":
           PGASetStoppingRuleType (self.ctx, params[0])
        elif item == "MaxGAIterValue":
           PGASetMaxGAIterValue(self.ctx, params[0])
        elif item == "MaxNoChangeValue":
           PGASetMaxNoChangeValue(self.ctx, params[0])
        elif item == "MaxSimilarityValue":
           PGASetMaxSimilarityValue(self.ctx, params[0])
#        elif item == "UserFunction":
#           constant, void *f= value
#           PGASetUserFunction(self.ctx, constant, void *f)

    def Get(self,item,*params):
#        return PGAGetDataType (self.ctx)
        if item == "BinaryAllele":
           return PGAGetBinaryAllele ( self.ctx, params[0], params[1], params[2])
        elif item == "BinaryInitProb":
           return PGAGetBinaryInitProb (self.ctx)
        elif item == "CharacterAllele":
           return PGAGetCharacterAllele (self.ctx, params[0], params[1], params[0])
        elif item == "RandomInitFlag":
           return PGAGetRandomInitFlag (self.ctx)
        elif item == "CrossoverType":
           return PGAGetCrossoverType (self.ctx)
        elif item == "CrossoverProb":
           return PGAGetCrossoverProb (self.ctx)
        elif item == "UniformCrossoverProb":
           return PGAGetUniformCrossoverProb (self.ctx)
#        elif item == "DebugLevelOfName":
#           return PGAGetDebugLevelOfName(self.ctx, *funcname)
#        elif item == "DebugFlag":
#           return PGAGetDebugFlag(self.ctx, *funcname)
        elif item == "NoDuplicatesFlag":
           return PGAGetNoDuplicatesFlag (self.ctx)
        elif item == "Evaluation":
           return PGAGetEvaluation ( self.ctx, params[0], params[1])
        elif item == "EvaluationUpToDateFlag":
           return PGAGetEvaluationUpToDateFlag ( self.ctx, params[0], params[1])
        elif item == "RealFromBinary":
           return PGAGetRealFromBinary(self.ctx, params[0], params[1], params[2], params[3], params[4], params[5])
        elif item == "RealFromGrayCode":
           return PGAGetRealFromGrayCode(self.ctx, params[0], params[1], params[2], params[3], params[4], params[5])
        elif item == "IntegerFromBinary":
           return PGAGetIntegerFromBinary(self.ctx, params[0], params[1], params[2], params[3])
        elif item == "IntegerFromGrayCode":
           return PGAGetIntegerFromGrayCode(self.ctx, params[0], params[1], params[2], params[3])
        elif item == "Fitness":
           return PGAGetFitness ( self.ctx, params[0], params[1] )
        elif item == "FitnessType":
           return PGAGetFitnessType (self.ctx)
        elif item == "FitnessMinType":
           return PGAGetFitnessMinType (self.ctx)
        elif item == "MaxFitnessRank":
           return PGAGetMaxFitnessRank (self.ctx)
        elif item == "FitnessCmaxValue":
           return PGAGetFitnessCmaxValue (self.ctx)
        elif item == "IntegerAllele":
           return PGAGetIntegerAllele (self.ctx, params[0], params[1], params[2])
        elif item == "IntegerInitType":
           return PGAGetIntegerInitType (self.ctx)
        elif item == "MinIntegerInitValue":
           return PGAGetMinIntegerInitValue (self.ctx, params[0])
        elif item == "MaxIntegerInitValue":
           return PGAGetMaxIntegerInitValue (self.ctx, params[0])
        elif item == "MutationType":
           return PGAGetMutationType (self.ctx)
        elif item == "MutationRealValue":
           return PGAGetMutationRealValue (self.ctx)
        elif item == "MutationIntegerValue":
           return PGAGetMutationIntegerValue (self.ctx)
        elif item == "MutationBoundedFlag":
           return PGAGetMutationBoundedFlag(self.ctx)
        elif item == "MutationProb":
           return PGAGetMutationProb (self.ctx)
##        elif item == "Rank":
##           return PGAGetRank (self.ctx, MPI_Comm comm)
##        elif item == "NumProcs":
##           return PGAGetNumProcs (self.ctx, MPI_Comm comm)
        elif item == "NumIslands":
           return PGAGetNumIslands (self.ctx)
        elif item == "NumDemes":
           return PGAGetNumDemes (self.ctx)
##        elif item == "Communicator":
##           return PGAGetCommunicator( self.ctx)
        elif item == "DataType":
           return PGAGetDataType (self.ctx)
        elif item == "OptDirFlag":
           return PGAGetOptDirFlag (self.ctx)
        elif item == "StringLength":
           return PGAGetStringLength (self.ctx)
        elif item == "VariableStringLength":
           return PGAGetVariableStringLength (self.ctx, params[0], params[1])
        elif item == "GAIterValue":
           return PGAGetGAIterValue (self.ctx)
        elif item == "MutationOrCrossoverFlag":
           return PGAGetMutationOrCrossoverFlag (self.ctx)
        elif item == "MutationAndCrossoverFlag":
           return PGAGetMutationAndCrossoverFlag (self.ctx)
        elif item == "PopSize":
           return PGAGetPopSize (self.ctx)
        elif item == "NumReplaceValue":
           return PGAGetNumReplaceValue (self.ctx)
        elif item == "PopReplaceType":
           return PGAGetPopReplaceType (self.ctx)
        elif item == "SortedPopIndex":
           return PGAGetSortedPopIndex ( self.ctx, params[0])
        elif item == "RandomSeed":
           return PGAGetRandomSeed(self.ctx)
        elif item == "RealAllele":
           return PGAGetRealAllele (self.ctx, params[0], params[1], params[2])
        elif item == "MinRealInitValue":
           return PGAGetMinRealInitValue (self.ctx, params[0])
        elif item == "MaxRealInitValue":
           return PGAGetMaxRealInitValue (self.ctx, params[0])
        elif item == "RealInitType":
           return PGAGetRealInitType (self.ctx)
        elif item == "PrintFrequencyValue":
           return PGAGetPrintFrequencyValue (self.ctx)
        elif item == "RestartFlag":
           return PGAGetRestartFlag(self.ctx)
        elif item == "RestartFrequencyValue":
           return PGAGetRestartFrequencyValue(self.ctx)
        elif item == "RestartAlleleChangeProb":
           return PGAGetRestartAlleleChangeProb(self.ctx)
        elif item == "SelectType":
           return PGAGetSelectType (self.ctx)
        elif item == "PTournamentProb":
           return PGAGetPTournamentProb(self.ctx)
        elif item == "StoppingRuleType":
           return PGAGetStoppingRuleType (self.ctx)
        elif item == "MaxGAIterValue":
           return PGAGetMaxGAIterValue (self.ctx)
#        elif item == "MaxMachineIntValue":
#           return PGAGetMaxMachineIntValue (self.ctx)
#        elif item == "MinMachineIntValue":
#           return PGAGetMinMachineIntValue (self.ctx)
#        elif item == "MaxMachineDoubleValue":
#           return PGAGetMaxMachineDoubleValue (self.ctx)
#        elif item == "MinMachineDoubleValue":
#           return PGAGetMinMachineDoubleValue (self.ctx)
        elif item == "WorstIndex":
           return PGAGetWorstIndex(self.ctx, params[0])
        elif item == "BestIndex":
           return PGAGetBestIndex(self.ctx, params[0])
##        elif item == "Individual":
##           p, pop= params
##           return PGAGetIndividual ( self.ctx, p, pop)



